"""hikari - software for handling and analysing crystallographic files.
Copyright (c) 2021, Daniel Tchoń.
All rights reserved.

This source code is licensed under the MIT-style license found in the
LICENSE file in the root directory of this source tree."""

MEMORY_SIZE = 500 * 2 ** 20
"""Approximate upper limit of taken memory in bytes"""
